#include "Win32DxInterface.h"

#include "FileLogger.h"

#include <DirectXMath.h>
#include <DirectXPackedVector.h>

#include <dxgi.h>
#include <dxgi1_6.h>
#include <d3d11.h>
#include <d3d11on12.h>
#include <d3d12.h>
#include <d3d12sdklayers.h>

#include <d2d1_3.h>
#include <d2d1_1.h>

#include <dwrite.h>

#include <iostream>
#include <vector>

#include <windows.h>
#include <wrl.h>

void Win32DxInterface::getHardwareAdapter(IDXGIAdapter1** ppAdapter)
{
	*ppAdapter = nullptr;
	for (UINT adapterIndex = 0; ; ++adapterIndex)
	{
		IDXGIAdapter1* pAdapter = nullptr;
		if (DXGI_ERROR_NOT_FOUND == mFactory->EnumAdapters1(adapterIndex, &pAdapter))
		{
			break;
		}

		// Check to see if the adapter supports Direct3D 12, but don't create the actual device yet.
		if (SUCCEEDED(D3D12CreateDevice(pAdapter, D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr)))
		{
			*ppAdapter = pAdapter;
			return;
		}
		pAdapter->Release();
	}
}

bool Win32DxInterface::isValid() const
{
	return mIsValid;
}

ID3D12Device* Win32DxInterface::getDevice() const
{
	return mDevice.Get();
}

ID3D11On12Device* Win32DxInterface::get11On12Device() const
{
	return mD3d11On12Device.Get();
}

IDXGIFactory7* Win32DxInterface::getFactory() const
{
	return mFactory.Get();
}

ID2D1Factory3* Win32DxInterface::getD2dFactory() const
{
	return mD2dFactory.Get();
}

IDWriteFactory* Win32DxInterface::getDirectWriteFactory() const
{
	return mDWriteFactory.Get();
}

ID2D1DeviceContext2* Win32DxInterface::getD2dContext() const
{
	return mD2dDeviceContext.Get();
}

ID3D12CommandQueue* Win32DxInterface::getCommandQueue() const
{
	return mCommandQueue.Get();
}

ID3D11DeviceContext* Win32DxInterface::getD3d11DeviceContext() const
{
	return mD3d11DeviceContext.Get();
}

void Win32DxInterface::initialize()
{
	MLOG_INFO("Initialize DX");
	UINT dxgiFactoryFlags = 0;
#if defined(DEBUG) || defined(_DEBUG)
	{
		Microsoft::WRL::ComPtr<ID3D12Debug> debugController;
		if (SUCCEEDED(::D3D12GetDebugInterface(IID_PPV_ARGS(&debugController))))
		{
			debugController->EnableDebugLayer();
			dxgiFactoryFlags |= DXGI_CREATE_FACTORY_DEBUG;
		}
	}
#endif

	HRESULT hr = ::CreateDXGIFactory2(dxgiFactoryFlags, IID_PPV_ARGS(&mFactory));

	Microsoft::WRL::ComPtr<IDXGIAdapter1> pAdapter;
	getHardwareAdapter(&pAdapter);
	if (pAdapter)
	{
		if (!SUCCEEDED(D3D12CreateDevice(pAdapter.Get(), D3D_FEATURE_LEVEL_11_0, IID_PPV_ARGS(&mDevice))))
		{
			MLOG_ERROR("Failed to create D3D12 device.");
			return;
		}
	}
	else
	{
		MLOG_ERROR("Failed to get DX hardware adapter.");
		return;
	}

	D3D12_COMMAND_QUEUE_DESC queueDesc = {};
	queueDesc.Flags = D3D12_COMMAND_QUEUE_FLAG_NONE;
	queueDesc.Type = D3D12_COMMAND_LIST_TYPE_DIRECT;
	if (!SUCCEEDED(mDevice->CreateCommandQueue(&queueDesc, IID_PPV_ARGS(&mCommandQueue))))
	{
		MLOG_ERROR("Failed to create command queue.");
		return;
	}

	const auto init_2d_ok = initializeD2d();
	mIsValid = init_2d_ok;
}

bool Win32DxInterface::initializeD2d()
{
	if (!initializeD11on12())
	{
		return false;
	}

	D2D1_FACTORY_OPTIONS d2dFactoryOptions = {};
	D2D1CreateFactory(D2D1_FACTORY_TYPE_SINGLE_THREADED, __uuidof(ID2D1Factory3), &d2dFactoryOptions, &mD2dFactory);
	Microsoft::WRL::ComPtr<IDXGIDevice> dxgiDevice;
	mD3d11On12Device.As(&dxgiDevice);

	mD2dFactory->CreateDevice(dxgiDevice.Get(), &mD2dDevice);

	D2D1_DEVICE_CONTEXT_OPTIONS deviceOptions = D2D1_DEVICE_CONTEXT_OPTIONS_NONE;
	mD2dDevice->CreateDeviceContext(deviceOptions, &mD2dDeviceContext);

	return initializeDirectWrite();
}

bool Win32DxInterface::initializeD11on12()
{
	Microsoft::WRL::ComPtr<ID3D11Device> d3d11Device;
	UINT d3d11DeviceFlags = D3D11_CREATE_DEVICE_BGRA_SUPPORT;
	D3D11On12CreateDevice(
		mDevice.Get(),
		d3d11DeviceFlags,
		nullptr,
		0,
		reinterpret_cast<IUnknown**>(mCommandQueue.GetAddressOf()),
		1,
		0,
		&d3d11Device,
		&mD3d11DeviceContext,
		nullptr
	);
	d3d11Device.As(&mD3d11On12Device);
	return true;
}

bool Win32DxInterface::initializeDirectWrite()
{
	DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory), &mDWriteFactory);
	return true;
}