#include "Win32DxWindowInterface.h"

#include "Win32DxInterface.h"
#include "Win32Window.h"
#include "Window.h"
#include "Widget.h"
#include "FileLogger.h"

#include "DrawingContext.h"
#include "DirectXPainter.h"
#include "DirectXMeshPainter.h"
#include "DirectXTextPainter.h"
#include "DirectXShaderProgram.h"

#include <dxgi.h>
#include <dxgi1_6.h>
#include <d3dcompiler.h>
#include <d3d12sdklayers.h>

#include <d3d11.h>
#include <d3d11on12.h>
#include <d2d1_3.h>
#include <d2d1_1.h>
#include <D3Dcommon.h>

#include <dwrite.h>

Win32DxWindowInterface::Win32DxWindowInterface(mt::Window* window, Win32DxInterface* dxInterface)
	:   mWindow(window),
        mDxInterface(dxInterface)
{

}

Win32DxWindowInterface::~Win32DxWindowInterface()
{

}

void Win32DxWindowInterface::destroyWindow()
{
    waitForPreviousFrame();

    ::CloseHandle(mFenceEvent);
}

void Win32DxWindowInterface::onRender()
{
    if (!mIsValid)
    {
        return;
    }

    // Record all the commands we need to render the scene into the command list.
    populateCommandList();

    // Execute the command list.
    ID3D12CommandList* ppCommandLists[] = { mCommandList.Get() };
    mDxInterface->getCommandQueue()->ExecuteCommandLists(_countof(ppCommandLists), ppCommandLists);

    renderD2d();

    // Present the frame.
    mSwapChain->Present(1, 0);

    waitForPreviousFrame();
}

void Win32DxWindowInterface::populateCommandList()
{
    auto painter = dynamic_cast<DirectXPainter*>(mWindow->getDrawingContent()->getPainter());

    // Command list allocators can only be reset when the associated 
    // command lists have finished execution on the GPU; apps should use 
    // fences to determine GPU execution progress.
    mCommandAllocator->Reset();

    // However, when ExecuteCommandList() is called on a particular command 
    // list, that command list can then be reset at any time and must be before 
    // re-recording.
    mCommandList->Reset(mCommandAllocator.Get(), painter->getMeshPainter()->getPipelineState());

    // Set necessary state.
    mCommandList->SetGraphicsRootSignature(painter->getMeshPainter()->getRootSignature());
    mCommandList->RSSetViewports(1, &mViewport);
    mCommandList->RSSetScissorRects(1, &mScissorRect);

    // Indicate that the back buffer will be used as a render target.
    auto barrier = CD3DX12_RESOURCE_BARRIER::Transition(mRenderTargets[mFrameIndex].Get(), D3D12_RESOURCE_STATE_PRESENT, D3D12_RESOURCE_STATE_RENDER_TARGET);
    mCommandList->ResourceBarrier(1, &barrier);

    CD3DX12_CPU_DESCRIPTOR_HANDLE rtvHandle(mRtvHeap->GetCPUDescriptorHandleForHeapStart(), static_cast<INT>(mFrameIndex), mRtvDescriptorSize);
    mCommandList->OMSetRenderTargets(1, &rtvHandle, FALSE, nullptr);

    painter->paintBackground(rtvHandle, mCommandList.Get());
    painter->paintMesh(mCommandList.Get());

    // Indicate that the back buffer will now be used to present.
    //barrier = CD3DX12_RESOURCE_BARRIER::Transition(mRenderTargets[mFrameIndex].Get(), D3D12_RESOURCE_STATE_RENDER_TARGET, D3D12_RESOURCE_STATE_PRESENT);
    //mCommandList->ResourceBarrier(1, &barrier);

    mCommandList->Close();
}

void Win32DxWindowInterface::renderD2d()
{
    // Acquire our wrapped render target resource for the current back buffer.
    mDxInterface->get11On12Device()->AcquireWrappedResources(mWrappedBackBuffers[mFrameIndex].GetAddressOf(), 1);

    // Render text directly to the back buffer.
    mDxInterface->getD2dContext()->SetTarget(mD2dRenderTargets[mFrameIndex].Get());

    auto painter = dynamic_cast<DirectXPainter*>(mWindow->getDrawingContent()->getPainter());
    painter->paintText(mDxInterface->getD2dContext(), mDxInterface->getDirectWriteFactory());

    // Release our wrapped render target resource. Releasing 
    // transitions the back buffer resource to the state specified
    // as the OutState when the wrapped resource was created.
    mDxInterface->get11On12Device()->ReleaseWrappedResources(mWrappedBackBuffers[mFrameIndex].GetAddressOf(), 1);

    // Flush to submit the 11 command list to the shared command queue.
    mDxInterface->getD3d11DeviceContext()->Flush();
}

bool Win32DxWindowInterface::initialize()
{
	if (mInitialized)
	{
		return true;
	}
    mInitialized = true;

    const auto width = mWindow->getWidth();
    const auto height = mWindow->getHeight();
    mViewport = CD3DX12_VIEWPORT(0.0f, 0.0f, static_cast<float>(width), static_cast<float>(height));
    mScissorRect = CD3DX12_RECT(0, 0, width, height);

    if (!loadPipeline())
    {
        return false;
    }

    if (!loadAssets())
    {
        return false;
    }

    mIsValid = true;
    return true;
}

bool Win32DxWindowInterface::loadPipeline()
{
    if (!setupSwapChain())
    {
        return false;
    }

    setupDescriptorHeaps();

    setupFrameResources();

    mDxInterface->getDevice()->CreateCommandAllocator(D3D12_COMMAND_LIST_TYPE_DIRECT, IID_PPV_ARGS(&mCommandAllocator));

    return true;
}

bool Win32DxWindowInterface::setupSwapChain()
{
    const auto width = mWindow->getWidth();
    const auto height = mWindow->getHeight();
    const auto hwnd = dynamic_cast<Win32Window*>(mWindow->getPlatformWindow())->getHandle();

    DXGI_SWAP_CHAIN_DESC1 swapChainDesc = {};
    swapChainDesc.BufferCount = FrameCount;
    swapChainDesc.Width = width;
    swapChainDesc.Height = height;
    swapChainDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
    swapChainDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
    swapChainDesc.SwapEffect = DXGI_SWAP_EFFECT_FLIP_DISCARD;
    swapChainDesc.SampleDesc.Count = 1;

    auto command_queue = mDxInterface->getCommandQueue();
    Microsoft::WRL::ComPtr<IDXGISwapChain1> swapChain;
    const auto hr = mDxInterface->getFactory()->CreateSwapChainForHwnd(command_queue, hwnd, &swapChainDesc, nullptr, nullptr, &swapChain);
    if (!SUCCEEDED(hr))
    {
        MLOG_ERROR("Failed to create swap chain for hwnd: " << hwnd);
        return false;
    }
    swapChain.As(&mSwapChain);

    // Prevent fullscreen toggle
    mDxInterface->getFactory()->MakeWindowAssociation(hwnd, DXGI_MWA_NO_ALT_ENTER);
    
    // Update index for backbuffer
    mFrameIndex = mSwapChain->GetCurrentBackBufferIndex();
    return true;
}

void Win32DxWindowInterface::setupDescriptorHeaps()
{
    D3D12_DESCRIPTOR_HEAP_DESC rtvHeapDesc = {};
    rtvHeapDesc.NumDescriptors = FrameCount;
    rtvHeapDesc.Type = D3D12_DESCRIPTOR_HEAP_TYPE_RTV;
    rtvHeapDesc.Flags = D3D12_DESCRIPTOR_HEAP_FLAG_NONE;
    mDxInterface->getDevice()->CreateDescriptorHeap(&rtvHeapDesc, IID_PPV_ARGS(&mRtvHeap));

    D3D12_DESCRIPTOR_HEAP_DESC srvHeapDesc = {};
    srvHeapDesc.NumDescriptors = 1;
    srvHeapDesc.Type = D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV;
    srvHeapDesc.Flags = D3D12_DESCRIPTOR_HEAP_FLAG_SHADER_VISIBLE;
    mDxInterface->getDevice()->CreateDescriptorHeap(&srvHeapDesc, IID_PPV_ARGS(&mSrvHeap));

    D3D12_DESCRIPTOR_HEAP_DESC cbvHeapDesc = {};
    cbvHeapDesc.NumDescriptors = 1;
    cbvHeapDesc.Type = D3D12_DESCRIPTOR_HEAP_TYPE_CBV_SRV_UAV;
    cbvHeapDesc.Flags = D3D12_DESCRIPTOR_HEAP_FLAG_SHADER_VISIBLE;
    mDxInterface->getDevice()->CreateDescriptorHeap(&cbvHeapDesc, IID_PPV_ARGS(&mCbvHeap));

    mRtvDescriptorSize = mDxInterface->getDevice()->GetDescriptorHandleIncrementSize(D3D12_DESCRIPTOR_HEAP_TYPE_RTV);
}

void Win32DxWindowInterface::setupFrameResources()
{
    CD3DX12_CPU_DESCRIPTOR_HANDLE rtvHandle(mRtvHeap->GetCPUDescriptorHandleForHeapStart());

    float dpiX{ 0.0 };
    float dpiY{ 0.0 };
    mDxInterface->getD2dFactory()->GetDesktopDpi(&dpiX, &dpiY);
    D2D1_BITMAP_PROPERTIES1 bitmapProperties = D2D1::BitmapProperties1(
        D2D1_BITMAP_OPTIONS_TARGET | D2D1_BITMAP_OPTIONS_CANNOT_DRAW,
        D2D1::PixelFormat(DXGI_FORMAT_UNKNOWN, D2D1_ALPHA_MODE_PREMULTIPLIED),
        dpiX,
        dpiY
    );

    for (UINT n = 0; n < FrameCount; n++)
    {
        setupFrameResource(n, rtvHandle, bitmapProperties);
    }
}

void Win32DxWindowInterface::setupFrameResource(UINT idx, CD3DX12_CPU_DESCRIPTOR_HANDLE& descriptorHandle, const D2D1_BITMAP_PROPERTIES1& bitmapProps)
{
    mSwapChain->GetBuffer(idx, IID_PPV_ARGS(&mRenderTargets[idx]));
    mDxInterface->getDevice()->CreateRenderTargetView(mRenderTargets[idx].Get(), nullptr, descriptorHandle);

    setupFrameD2DResource(idx, bitmapProps);

    descriptorHandle.Offset(1, mRtvDescriptorSize);
}

void Win32DxWindowInterface::setupFrameD2DResource(UINT idx, const D2D1_BITMAP_PROPERTIES1& bitmapProps)
{
    D3D11_RESOURCE_FLAGS d3d11Flags = { D3D11_BIND_RENDER_TARGET };
    mDxInterface->get11On12Device()->CreateWrappedResource(mRenderTargets[idx].Get(),&d3d11Flags,
        D3D12_RESOURCE_STATE_RENDER_TARGET, D3D12_RESOURCE_STATE_PRESENT, IID_PPV_ARGS(&mWrappedBackBuffers[idx]));

    Microsoft::WRL::ComPtr<IDXGISurface> surface;
    mWrappedBackBuffers[idx].As(&surface);
    mDxInterface->getD2dContext()->CreateBitmapFromDxgiSurface(surface.Get(), &bitmapProps, &mD2dRenderTargets[idx]);
}

bool Win32DxWindowInterface::loadAssets()
{
    auto painter = dynamic_cast<DirectXPainter*>(mWindow->getDrawingContent()->getPainter());

    painter->initializeMesh(mDxInterface->getDevice());

    // Create the command list.
    mDxInterface->getDevice()->CreateCommandList(0, D3D12_COMMAND_LIST_TYPE_DIRECT, mCommandAllocator.Get(), 
        painter->getMeshPainter()->getPipelineState(), IID_PPV_ARGS(&mCommandList));

    // Create D2D/DWrite objects for rendering text.
    painter->initializeText(mDxInterface->getD2dContext(), mDxInterface->getDirectWriteFactory());

    // Command lists are created in the recording state, but there is nothing
    // to record yet. The main loop expects it to be closed, so close it now.
    mCommandList->Close();

    painter->updateMesh(mDxInterface->getDevice());

    createSyncObjects();
    return true;
}

void Win32DxWindowInterface::createSyncObjects()
{
    mDxInterface->getDevice()->CreateFence(0, D3D12_FENCE_FLAG_NONE, IID_PPV_ARGS(&mFence));
    mFenceValue = 1;

    // Create an event handle to use for frame synchronization.
    mFenceEvent = ::CreateEvent(nullptr, FALSE, FALSE, nullptr);
    if (mFenceEvent == nullptr)
    {
        //ThrowIfFailed(HRESULT_FROM_WIN32(GetLastError()));
    }

    // Wait for the command list to execute; we are reusing the same command 
    // list in our main loop but for now, we just want to wait for setup to 
    // complete before continuing.
    waitForPreviousFrame();
}

void Win32DxWindowInterface::waitForPreviousFrame()
{
    // Signal and increment the fence value.
    const UINT64 fence = mFenceValue;
    mDxInterface->getCommandQueue()->Signal(mFence.Get(), fence);
    mFenceValue++;

    // Wait until the previous frame is finished.
    if (mFence->GetCompletedValue() < fence)
    {
        mFence->SetEventOnCompletion(fence, mFenceEvent);
        ::WaitForSingleObject(mFenceEvent, INFINITE);
    }
    mFrameIndex = mSwapChain->GetCurrentBackBufferIndex();
}